local lib = LibIJA
local internal = lib.internal

-------------------------------------
-- Sort Functions
-------------------------------------
local SORT_KEYS ={
	maxPrice = { tiebreaker = "minPrice", isNumeric = true},
	minPrice = { tiebreaker = "numReagents" , isNumeric = true, tieBreakerSortOrder = ZO_SORT_ORDER_DOWN},
	totalPrice = { tiebreaker = "numReagents" , isNumeric = true, tieBreakerSortOrder = ZO_SORT_ORDER_DOWN},
	sellPrice = { tiebreaker = "numReagents" , isNumeric = true, tieBreakerSortOrder = ZO_SORT_ORDER_DOWN},
	
	priority = { tiebreaker = "numReagents" , isNumeric = true, tieBreakerSortOrder = ZO_SORT_ORDER_DOWN},
	
	numReagents = { tiebreaker = "stackCount" , isNumeric = true},
	stackCount = { tiebreaker = "itemId" , isNumeric = true, tieBreakerSortOrder = ZO_SORT_ORDER_UP},
	itemId = { tiebreaker = "name" , isNumeric = true},
	name = { }
}
local STACK_SORT_KEYS ={
	numReagents = { tiebreaker = "stackCount", isNumeric = true, tieBreakerSortOrder = ZO_SORT_ORDER_DOWN },
	stackCount = { tiebreaker = "itemId" , isNumeric = true, tieBreakerSortOrder = ZO_SORT_ORDER_UP},
	itemId = { tiebreaker = "name" , isNumeric = true},
	name = { }
}

local function PatternSortByMaxPrice(data1, data2)
	return ZO_TableOrderingFunction(data1, data2, "maxPrice", SORT_KEYS, ZO_SORT_ORDER_UP)
end
local function PatternSortByValue(data1, data2)
	return ZO_TableOrderingFunction(data1, data2, "totalPrice", SORT_KEYS, ZO_SORT_ORDER_UP)
end

local function SortByStock(data1, data2)
	return ZO_TableOrderingFunction(data1, data2, "stackCount", SORT_KEYS, ZO_SORT_ORDER_DOWN)
end
--[[
local function PatternSortByStock(entry1, entry2)
	return ZO_TableOrderingFunction(entry1, entry2, "numReagents", SORT_KEYS, ZO_SORT_ORDER_UP)
end
--]]
local function PatternSortByStock(a, b)
	local numA = #a.reagents
	local numB = #b.reagents
	
	if numA ~= numB then
		return numA < numB
	end

	local minStackA = a:GetMinStock()
	local minStackB = b:GetMinStock()
	
	if minStackA ~= minStackB then
		return minStackA > minStackB
	end
	local maxStackA	= a:GetMaxStock()
	local maxStackB	= b:GetMaxStock()
	
	if maxStackA ~= maxStackB then
		return maxStackA > maxStackB
	end
	
	return ZO_TableOrderingFunction(a, b, "itemId", SORT_KEYS, ZO_SORT_ORDER_DOWN)
end

local function SortBySellPrice(data1, data2)
	return ZO_TableOrderingFunction(data1, data2, "sellPrice", SORT_KEYS, ZO_SORT_ORDER_UP)
end
local function SortByManual(data1, data2)
	return ZO_TableOrderingFunction(data1, data2, "priority", SORT_KEYS, ZO_SORT_ORDER_UP)
end
local function SortByID(data1, data2)
	return ZO_TableOrderingFunction(data1, data2, "itemId", SORT_KEYS, ZO_SORT_ORDER_UP)
end

local function PatternSortByTraits(a, b)
	local traitIdA1 = a.traits[1]
	local traitIdA2 = a.traits[2] or 0
	local traitIdA3 = a.traits[3] or traitIdA2
	
	local traitIdB1 = b.traits[1]
	local traitIdB2 = b.traits[2] or 0
	local traitIdB3 = b.traits[3] or traitIdB2
	
	
	
	if traitIdA1 ~= traitIdB1 then
		return traitIdA1 < traitIdB1
	end

	if not traitIdA2 or traitIdB2 then return false end
	
	if traitIdA2 ~= traitIdB2 then
		return traitIdA2 < traitIdB2
	end
	
	if not traitIdA3 or traitIdB3 then return false end
	return traitIdA3 < traitIdB3
	
end

-------------------------------------
-- 
-------------------------------------
local function getDummyData(reagent)
	local linkFormat = "|H0:item:<<1>>:30:1:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0|h|h"
	local itemLink = zo_strformat(linkFormat, reagent.itemId)
	local icon, _, meetsUsageRequirement = GetItemLinkInfo(itemLink)
	local data = {
		name = reagent.name,
		itemLink = itemLink,
		icon = icon,
		stackCount = 0
	}
	return data
end

local function getTraitPriorities(traits)
 	local traitPrioritys = {}
	for i=1, 4 do
		traitPrioritys[#traitPrioritys + 1] = traits[i]
	end
	return traitPrioritys
end

-------------------------------------
-- 
-------------------------------------
local Recipe_Data = ZO_InitializingObject:Subclass()

function Recipe_Data:Initialize()
	self.reagents = {}
end

function Recipe_Data:AddAdditionalData(numRecipes, index)
	local solvent 		= ZO_ShallowTableCopy(lib.currentData.solvent)
	self.solventData 	= solvent
	self.resultLink 	= lib:GetAlchemyResultLink(lib.currentData.type, lib.currentData.traitIds, solvent.itemId)
	self.variations		= zo_strformat(GetString(SI_LIBIJA_THIS_OF_THAT), index, numRecipes)
	self.index			= index
	self.iterations		= 1	-- todo - see about removing
end

function Recipe_Data:BuildData(reagents, combinedTraits, traitKeys, traitPrioritys)
	self.reagents = {}
	self:SetReagentData( reagents )
	self:SortReagents(self.reagents)
	
	self.traits			= combinedTraits 
	
	self.traitKeys		= traitKeys
	self.numTraits		= #combinedTraits
	self.traitPrioritys	= traitPrioritys

	-- sort keys
	self.numReagents	= #reagents
	
	self.totalPrice		= self:GetTotalPrice()
	self.maxPrice		= self:GetMaxPrices()
	self.minPrice		= self:GetMinPrices()
	self.priority		= self:GetRecipePriority()
	
	self:GetAveragePrices()
end

function Recipe_Data:GetReagentData(reagent)
 	local itemId = reagent.itemId
	local reagentData = lib:GetItemData(itemId, IJA_IsAlchemyReagent) or getDummyData(reagent)
	local itemLink = reagentData.itemLink
	
	reagentData.traits 			= reagent.traits
	reagentData.priority		= self:GetPriority(itemId)
	reagentData.traitPrioritys	= getTraitPriorities(reagent.traits)
	reagentData.sellPrice		= reagentData.sellPrice
--	reagentData.sellPrice		= internal:GetAvgPrice(itemLink)
	
	return reagentData
end

-------------------------------------
function Recipe_Data:GetPriority(itemId)
     if LibIJA_Data.priorityBy ~= IJA_REAGENT_PRIORITY_BY_MANUAL then
 --       return nil
    end
    if (not itemId) or (itemId == 0) then
        return #LibIJA_Data.priorityByManual + 1
    end

    for key, value in ipairs(LibIJA_Data.priorityByManual) do
        if value == itemId then
            return key
        end
    end
    return #LibIJA_Data.priorityByManual + 1
end

function Recipe_Data:GetRecipePriority()
	local priority = 0
	for k,v in pairs(self.reagents) do
		priority = v.priority > priority and v.priority or priority
	end
	return priority
end

-------------------------------------
function Recipe_Data:GetTotalStock()
	local total = 0
	for k, reagent in pairs(self.reagents) do
		total = total + reagent.stackCount
	end
	return total
end

function Recipe_Data:GetMaxStock()
	local price1 = self.reagents[1].stackCount
	local price2 = self.reagents[2].stackCount
	local price3 = self.reagents[3] and self.reagents[3].stackCount or 0
	
	return math.max(price1, price2, price3)
end

function Recipe_Data:GetMinStock()
	local price1 = self.reagents[1].stackCount
	local price2 = self.reagents[2].stackCount
	local price3 = self.reagents[3] and self.reagents[3].stackCount
	
	if price3 then
		return math.min(price1, price2, price3)
	else
		return math.min(price1, price2)
	end
end

-------------------------------------
function Recipe_Data:GetTotalPrice()
	local total = 0
	for k, reagent in pairs(self.reagents) do
		total = total + reagent.sellPrice
	end
	return total
end

function Recipe_Data:GetMaxPrices()
	local price1 = self.reagents[1].sellPrice
	local price2 = self.reagents[2].sellPrice
	local price3 = self.reagents[3] and self.reagents[3].sellPrice or 0
	
	return math.max(price1, price2, price3)
end

function Recipe_Data:GetMinPrices()
	local price1 = self.reagents[1].sellPrice
	local price2 = self.reagents[2].sellPrice
	local price3 = self.reagents[3] and self.reagents[3].sellPrice or 0
	
	return math.min(price1, price2, price3)
end

function Recipe_Data:GetAveragePrices()
	if TamrielTradeCentre then
		for k, reagent in pairs(self.reagents) do
			local priceInfo = TamrielTradeCentrePrice:GetPriceInfo(itemLink)
			if priceInfo and priceInfo.Avg then
				reagent.sellPrice = priceInfo.Avg or reagent.sellPrice
			end
		end
	end
end

-------------------------------------
function Recipe_Data:GetNumReagents()
	return #self.reagents
end

function Recipe_Data:SetReagentData( reagents )
	for k, reagent in pairs(reagents) do
		local reagentData = self:GetReagentData(reagent)
		table.insert(self.reagents, reagentData)
	end
end

function Recipe_Data:SortReagents(reagents)
 	local sortFunctions = {
		[IJA_REAGENT_PRIORITY_BY_STOCK] = SortByStock,
		[IJA_REAGENT_PRIORITY_BY_MANUAL] = SortByManual,
		[IJA_REAGENT_PRIORITY_BY_VALUE] = SortBySellPrice,
		[IJA_REAGENT_PRIORITY_BY_PRICE] = SortBySellPrice,
		[IJA_REAGENT_PRIORITY_BY_ID] = SortByID,
	}
	table.sort(reagents, sortFunctions[LibIJA_Data.priorityBy])
end

function Recipe_Data:Reset()
-- place holder
end

-------------------------------------
-- Lookup Functions
-------------------------------------
local resultKeys = {}

local function isNewMatch(key_1, key_2, key_3)
	local isNew = true
	local combination = key_1 * key_2
	
	if key_3 ~= nil then
		combination = combination * key_3
		if resultKeys[combination] then
			return false
		else
			resultKeys[combination] = true
			return true
		end
	else
		if resultKeys[combination] then
			return false
		else
			resultKeys[combination] = true
			return true
		end
	end
end

local function doesMatch(traits, traitKeys)
	for i=1, #traits do
		if traits[i] ~= 0 then
			if traitKeys and traitKeys[traits[i]] == nil then
				return false
			end
		end
	end
	return true
end

local function useThirdSlot()
	return LibIJA_Data.priorityBy == IJA_REAGENT_PRIORITY_BY_ID or ZO_Alchemy_IsThirdAlchemySlotUnlocked()
end

local function FindMatchingReagents(traits)
 	local reagents = {}
	for itemId, reagentData in pairs(internal.REAGENT_DATA) do
		for k, trait in pairs(traits) do
			if reagentData.traits[trait] then
				reagents[itemId] = reagentData
			end
		end		
	end
	return reagents
end

--[[
local function CombineReagentTraits(reagents)
	local combinedTraits, traitKeys, traitCounts, matches = {}, {}, {}, false

	if #reagents == 0 then return end
	for itemId, reagent in pairs(reagents) do
		for traitId, _ in pairs(reagent.traits) do
			if traitCounts[traitId] == nil then traitCounts[traitId] = 0 end
			traitCounts[traitId] = traitCounts[traitId] + 1
		end
	end
	
	local traitPrioritys = {}
	for traitId, count in pairs(traitCounts) do
		local traitData = internal.POTIONTRAIT_DATA[traitId]
		if traitCounts[traitData.oppositeId] then
			count = count - traitCounts[traitData.oppositeId] -- removes opposite from table
		end
		
		if (count > 1) then 
			combinedTraits[#combinedTraits+1] = traitId
			traitKeys[traitId] = true
			table.insert(traitPrioritys, traitData.index)
		end
	end
	
	table.sort(traitPrioritys, function(a, b) return a < b end)
	
	return combinedTraits, traitKeys, traitPrioritys[1]
end
--]]

local function CombineReagentTraits(reagents)
	local combinedTraits = lib:GetCombineReagentTraits(reagents)
	local traitKeys = {}
	local traitPrioritys = {}
	for _, traitId in pairs(combinedTraits) do
		local traitData = internal.POTIONTRAIT_DATA[traitId]
		traitKeys[traitId] = true
		table.insert(traitPrioritys, traitData.index)
	end
	
	table.sort(traitPrioritys, function(a, b) return a < b end)
	
	return combinedTraits, traitKeys, traitPrioritys[1]
end

local Search_Results = ZO_InitializingObject:Subclass()

function Search_Results:Initialize()
	self.async, self.scan, self.build = internal.async, internal.scan, internal.build
	
	local function CreateRecipeData()
		return Recipe_Data:New()
	end
	local function ResetData(data)
		data:Reset()
	end
	
	self.recipeObjectPool = ZO_ObjectPool:New(CreateRecipeData, ResetData)
end

function Search_Results:Process(traits, reagents, exactMatch)
	local result = {reagents = {}}
	
	local matched = true
	local combinedTraits, traitKeys, traitPrioritys = CombineReagentTraits(reagents)

	matched = exactMatch == nil and matched or exactMatch(traits, combinedTraits)
	
	if matched then
		if doesMatch(traits, traitKeys) then
			local recipeData = self.recipeObjectPool:AcquireObject()
			recipeData:BuildData(reagents, combinedTraits, traitKeys, traitPrioritys)
			
			return recipeData
		end
	end
end

function Search_Results:ScanForRecipes(traits, taskFunction, exactMatch)
	self:Reset()
	
 	local matchingReagents = FindMatchingReagents(traits)
	local allReagents = internal.REAGENT_DATA
	local reagents, recipeData = {}, {}
	local trait_1 = traits[1]
	local labUse = true
	
	resultKeys = {}
	local isInitialScan = self.async:GetCurrent()
	if isInitialScan then self.build:Suspend() end

	self.scan:For (pairs(matchingReagents)):Do(function(id_1, r_1Data)
		if id_1 then
			reagents = {[1] = r_1Data}
			for id_2, r_2Data in pairs(matchingReagents) do
				if id_1 ~= id_2 and isNewMatch(id_1, id_2) then
					reagents[2] = r_2Data
					local results_1 = taskFunction(self:Process(traits, reagents, exactMatch), false)
					if results_1 then 
						table.insert(recipeData, results_1) 
					end
					
					if useThirdSlot() then
						for id_3, r_3Data in pairs(allReagents) do
							if id_3 ~= id_1 and id_3 ~= id_2 and isNewMatch(id_1, id_2, id_3) then
								reagents[3] = r_3Data
								local results_2 = taskFunction(self:Process(traits, reagents, exactMatch), true)
								if results_2 then
									table.insert(recipeData, results_2)
								end
							end
						end
					end
				end
			end
		end
	end):Then(function()
		if isInitialScan then 
			internal.recipeData = recipeData
			self.build:Resume()
		else
			self:Sort(recipeData)
			internal:FireCallbacks("IJA_OnScanComplete", recipeData)
		end
	end)
end

function Search_Results:Sort(recipeData)
	 local sortFunctions = {
		[ IJA_REAGENT_PRIORITY_BY_STOCK]		= PatternSortByStock,
		[ IJA_REAGENT_PRIORITY_BY_MANUAL]		= SortByManual,
		[ IJA_REAGENT_PRIORITY_BY_VALUE]		= SortBySellPrice,
		[ IJA_REAGENT_PRIORITY_BY_PRICE]		= PatternSortByMaxPrice,
		[ IJA_REAGENT_PRIORITY_BY_ID]			= PatternSortByTraits,
	}
	table.sort(recipeData, sortFunctions[LibIJA_Data.priorityBy])
end

function Search_Results:Reset()
	self.recipeObjectPool:ReleaseAllObjects()
end

-------------------------------------
-- 
-------------------------------------
IJA_ALCHEMY_RECIPE_SCAN = Search_Results:New()



--	/script LibIJA.internal:InitializeScan()
--	/script IJA_ALCHEMY_RECIPE_DATA:Reset()
--[[

/script IJA_ALCHEMY_RECIPE_DATA.recipeObjectPool:AcquireObject()
--]]